/* vim:set ts=2 sw=2 sts=2 expandtab */
/*jshint asi: true undef: true es5: true node: true devel: true
         forin: true latedef: false globalstrict: true */
/*global define: true */

'use strict';

// In computing, the term stream is used in a number of ways, in all cases
// referring to a sequence of elements made available over time.


// Let's create a very basic stream, sequence of numbers from 3 to 0 by
// defining function that calls back given argument with each element.

function stream(next) {
  var i = 3
  while (i >= 0) next(i--)
}

// We can print this basic stream by passing a function to it that will print
// each element it will be called with:

stream(function onEach(element) {
  console.log(element)
})
// 3
// 2
// 1
// 0

// Since we will be printing streams quite frequently it is good idea to create
// a function that will help us to do that.

function print(stream) {
  stream(function onEach(element) {
    console.log(element)    // Print each element of the sequence.
  })
}

print(stream)

// 3
// 2
// 1
// 0

// You might think that we could have used array for defining sequence of
// elements. You are probably right, because our very basic stream was sequence
// of elements, ignoring other important part of definition: "made available
// over time".
//
// Let's create stream this time that makes elements available every 20ms.

function Randomizer(min, max) {
  var delta = max - min
  return function stream(next) {
    setInterval(function generate() {
      // We yield random number in given range every 20ms.
      next(min + Math.round(Math.random() * delta))
    }, 20)
  }
}

// We have defined `Randomizer` function takes `min` and `max` numbers and
// returns stream of random numbers in that range that become available every
// 20ms.

// No we can create stream of random numbers in 0 - 100 range.
var numberStream = Randomizer(0, 100)
// And we print our stream :)
print(numberStream)

// 29
// 33
// 45
// ....

// Oops! Stream keep printing this numbers infinitely. That's because stream is
// infinite! This means there are two types of streams finite and infinite. The
// difference between two is that finite streams stop when they rich end of the
// sequence.

// Since at least some streams stop we need to be able to know when that
// happens. Lets create another, finite stream, that can notify listener that
// stream is stopped once no element is left in sequence.

function list() {
  var elements = Array.prototype.slice.call(arguments, 0)
  return function stream(next, stop) {
    var index = 0
    // Yield each element of this stream by calling `next`.
    while (index < elements.length) next(elements[index++])
    // Once all elements are finished we stop a stream.
    if (stop) stop()
  }
}

// In addition 

// End lets define our print function with infinity in mind.

function print(stream) {
  console.log(">>>")
  stream(function onElement(element) {
    console.log(element)
  }, function onStop(error) {
    if (error) console.error(error)
    console.log("<<<")
  })
}

isInfinite(list(1, 2)) // > false

// Now lets print stream of elements

print(list(1, 'a', {}, [1,2]))

// >>>
// 1
// a
// {}
// [ 1, 2 ]
// <<<

// Lets try to define what stream is:
//
// Stream is a function that represents sequence of elements available over
// time. Stream takes one or two function arguments and calls first one every
// time element of it's sequence becomes available. Streams MAY be finite and
// infinite. Finite are streams that take two arguments, and call second one
// after calling a first function with a last element of the sequence.
// Infinite streams MUST takes only one argument.

// Let's do something from real life now. Let's try to define stream of
// directory's all entries including nested ones.

// To do so first we need to define function for creating stream of directory
// entries:

var fs = require("fs")
function ls(path) {
  return function stream(next, stop) {
    fs.readdir(path, function onEntries(error, entries) {
      var entry
      // On error we stop a stream
      if (error) return stop(error)
      // On each entry we yield element
      while ((entry = entries.shift())) next(entry)
      // Once we yielded last element we end a stream.
      stop()
    })
  }
}

// Lets print stream of entries of the current directory:

print(ls('./'))

// >>> Finite
// .gitignore
// History.md
// package.json
// readme.js
// Readme.md
// streamer.js
// tests
// <<<

// Now lets create function that returns stream of `stat` for the given `path`.
// (This stream will have one element and it's basically a promise / deferred
// :)

function stat(target) {
  return function stream(next, stop) {
    // see http://nodejs.org/docs/v0.4.8/api/fs.html#fs.stat
    fs.stat(target, function onStat(error, stats) {
      if (error) return stop(error)
      stats.path = target
      next(stats)
      stop()
    })
  }
}

var path = require("path")
function dirs(target) {
  var paths = map(ls(target), path.join.bind(null, target))
  var stats = merge(streamer.map(paths, stat))
  var dirStats = filter(function (stats) { return stats.isDirectory() })
  return map(dirStats, function(stats) { return stats.path })
}

function lstree(target) {
  var entries = map(ls(target), path.join.bind(null, target))
  var nested = merge(map(dirs(target), lstree))
  return join(entries, nested)
}

streamer.print(lstree('./'))
